from pathlib import Path

import pandas as pd
from config.core import DATASET_DIR, config
from pipeline import formation_energy_pipe
from processing.data_manager import load_dataset, save_pipeline
from sklearn.model_selection import train_test_split


def run_training() -> None:
    """Train the model."""

    # read training data
    data = load_dataset(file_name=config.app_config.dft_calc_train)

    # divide train and test
    X_train, X_test, y_train, y_test = train_test_split(
        data[config.model_config.input_features],
        data[config.model_config.target],
        test_size=config.model_config.test_size,
        random_state=config.model_config.random_state,
    )

    test_set = pd.concat([X_test, y_test], axis=1)
    test_set.to_csv(Path(f"{DATASET_DIR}/{config.app_config.test_split}"))

    # fit model
    formation_energy_pipe.fit(X_train, y_train)

    # persist trained model
    save_pipeline(pipeline_to_persist=formation_energy_pipe)


if __name__ == "__main__":
    run_training()



from pathlib import Path
from typing import List

from pydantic import BaseModel
from strictyaml import YAML, load

import regression_model

# Project Directories
PACKAGE_ROOT = Path(regression_model.__file__).resolve().parent
ROOT = PACKAGE_ROOT.parent
CONFIG_FILE_PATH = PACKAGE_ROOT / "config.yml"
DATASET_DIR = PACKAGE_ROOT / "datasets"
TRAINED_MODEL_DIR = PACKAGE_ROOT / "trained_models"


class AppConfig(BaseModel):
    """
    Application-level config.
    """

    package_name: str
    pipeline_save_file: str
    dft_calc_train: str
    dft_calc_test: str
    dft_calc_test_validation: str
    elemental_prop: str
    shan_rad: str
    a_shan_rad_cols: List[str]
    b_shan_rad_cols: List[str]
    no_nulls_in_required_loc: int
    a_site_not_present_loc: int
    b_site_not_present_loc: int
    non_stoichiometric_loc: int
    replace_blank_col: str
    replace_blank_loc: int
    num_site_test_col: str
    num_site_test_loc: int
    populate_elem_col: str
    populate_elem_loc: int
    a_shannon_radius_col: str
    a_shannon_radius_loc: int
    b_shannon_radius_col: str
    b_shannon_radius_loc: int
    comp_avgd_cols: List[str]
    comp_avgd_loc: int
    comp_avgd_expected_vals: List[float]
    avg_maj_col: str
    diff_maj_col: str
    avg_maj_loc: int
    diff_maj_loc: int
    test_split: str


class ModelConfig(BaseModel):
    """
    All configuration relevant to model
    training and feature engineering.
    """

    target: str
    features: List[str]
    test_size: float
    random_state: int
    remove_comma: str
    col_names_no_symbol: List[str]
    site_names: List[str]
    prefixes: List[str]
    elemental_property_cols: List[str]
    a_site_elements: List[str]
    b_site_elements: List[str]
    no_nulls_allowed: List[str]
    input_features: List[str]
    max_cols: List[str]
    composition: str
    relevant_properties: List[str]
    a_site_cols: List[str]
    b_site_cols: List[str]


class Config(BaseModel):
    """Master config object."""

    app_config: AppConfig
    model_config: ModelConfig


def find_config_file() -> Path:
    """
    Locate the configuration file.
    """

    if CONFIG_FILE_PATH.is_file():
        return CONFIG_FILE_PATH
    raise Exception(f"Config not found at {CONFIG_FILE_PATH!r}")


def fetch_config_from_yaml(cfg_path: Path = None) -> YAML:
    """
    Parse YAML containing the package configuration.
    """

    if not cfg_path:
        cfg_path = find_config_file()

    if cfg_path:
        with open(cfg_path, "r") as conf_file:
            parsed_config = load(conf_file.read())
            return parsed_config
    raise OSError(f"Did not find config file at path: {cfg_path}")


def create_and_validate_config(parsed_config: YAML = None) -> Config:
    """
    Run validation on config values.
    """

    if parsed_config is None:
        parsed_config = fetch_config_from_yaml()

    # specify the data attribute from the strictyaml YAML type.
    _config = Config(
        app_config=AppConfig(**parsed_config.data),
        model_config=ModelConfig(**parsed_config.data),
    )

    return _config


config = create_and_validate_config()



from lightgbm import LGBMRegressor
from sklearn.pipeline import Pipeline

from regression_model.config.core import config
from regression_model.processing.data_manager import load_dataset
from regression_model.processing.features import (
    CompositionAveragedProperties,
    KeepDesiredFeatures,
    NumOfSites,
    PopulateElementalProp,
    PopulateMajorityIonProperties,
    ReplaceBlank,
    ShannonRadius,
)

elemental_properties = load_dataset(file_name=config.app_config.elemental_prop)
shan_rad = load_dataset(file_name=config.app_config.shan_rad)

a_shan_rad = shan_rad[config.app_config.a_shan_rad_cols].dropna()
b_shan_rad = shan_rad[config.app_config.b_shan_rad_cols].dropna()

formation_energy_pipe = Pipeline(
    [
        ("replace_blank", ReplaceBlank()),
        (
            "populate_elemental_prop",
            PopulateElementalProp(
                ep=elemental_properties,
                prefixes=config.model_config.prefixes,
                sites=config.model_config.site_names,
            ),
        ),
        ("count_num_sites", NumOfSites(sites=config.model_config.site_names)),
        (
            "add_shannon_radius",
            ShannonRadius(
                a_shan=a_shan_rad,
                b_shan=b_shan_rad,
                prefixes=config.model_config.prefixes,
                sites=config.model_config.site_names,
            ),
        ),
        (
            "add_structural_parameters",
            CompositionAveragedProperties(
                sites=config.model_config.site_names,
                ep=elemental_properties,
                prefixes=config.model_config.prefixes,
            ),
        ),
        (
            "populate_majority_ion_properties",
            PopulateMajorityIonProperties(
                ep=elemental_properties, cols=config.model_config.max_cols
            ),
        ),
        (
            "keep_desired_features",
            KeepDesiredFeatures(features=config.model_config.features),
        ),
        ("lgbm_regressor", LGBMRegressor()),
    ]
)



import re
from typing import List, Optional, Tuple

import numpy as np
import pandas as pd
from pydantic import BaseModel, ValidationError

from regression_model.config.core import config


def check_stoichiometry(composition):
    elem_counts = re.findall("[0-9]+", composition)
    elem_counts = [int(num) for num in elem_counts]
    return sum(elem_counts) == 40


def drop_na_inputs(*, input_data: pd.DataFrame) -> pd.DataFrame:
    """Check model inputs for na values and filter."""
    validated_data = input_data.copy()

    if validated_data[config.model_config.no_nulls_allowed].isnull().values.any():
        raise ValueError("Nulls present in required features.")

    for a_site in config.model_config.a_site_cols:
        elems = validated_data[a_site].dropna()

        if not (elems.isin(config.model_config.a_site_elements)).min():
            raise ValueError(
                f"Some elements in {a_site} is not present in the Shannon radius directory."
            )

    for b_site in config.model_config.b_site_cols:
        elems = validated_data[b_site].dropna()
        if not (elems.isin(config.model_config.b_site_elements)).min():
            raise ValueError(
                f"Some elements in {b_site} is not present in the Shannon radius directory."
            )

    if (
        not validated_data[config.model_config.composition]
        .apply(check_stoichiometry)
        .min()
    ):
        raise ValueError("Perovskite oxide is not stoichiometric.")

    validated_data = validated_data.fillna(0)

    return validated_data


def validate_inputs(*, input_data: pd.DataFrame) -> Tuple[pd.DataFrame, Optional[dict]]:
    """Check model inputs for unprocessable values."""

    relevant_data = input_data[config.model_config.input_features].copy()

    try:
        validated_data = drop_na_inputs(input_data=relevant_data)
        errors = None

        # replace numpy nans so that pydantic can validate
        MultiplePerovskiteOxideSchema(
            inputs=validated_data.replace({np.nan: None}).to_dict(orient="records")
        )
    except ValidationError as error:
        errors = error.json()

    return validated_data, errors


class PerovskiteOxideSchema(BaseModel):
    COMPOSITION: str
    A_SITE_1: str
    A_SITE_2: Optional[str]
    A_SITE_3: Optional[str]
    B_SITE_1: str
    B_SITE_2: Optional[str]
    B_SITE_3: Optional[str]
    NUM_ELEMS: int


class MultiplePerovskiteOxideSchema(BaseModel):
    inputs: List[PerovskiteOxideSchema]



import logging

from regression_model.config.core import PACKAGE_ROOT, config

logging.getLogger(config.app_config.package_name).addHandler(logging.NullHandler())


with open(PACKAGE_ROOT / "VERSION") as version_file:
    __version__ = version_file.read().strip()



import re
from typing import List

import numpy as np
import numpy.typing as npt
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin

from regression_model.config.core import config


class ReplaceBlank(BaseEstimator, TransformerMixin):
    def __init__(self):
        return None

    def fit(self, X: pd.DataFrame, y: pd.Series = None):
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        X = X.copy()
        X = X.replace(" ", 0)
        X = X.fillna(0)
        return X


class NumOfSites(BaseEstimator, TransformerMixin):
    def __init__(self, sites: List[str]):
        self.sites = sites
        return None

    def fit(self, X: pd.DataFrame, y: pd.Series = None):
        return self

    def count_site_nums(self, composition: str, site: str) -> int:
        matches = re.findall(r"(\D*)(\d*)", composition)
        number_of_matches = len(matches)
        if site is not None:
            for j in range(number_of_matches):
                if matches[j][0] == site:
                    return int(matches[j][1])
                else:
                    continue
        return 0

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        X = X.copy()
        for site in self.sites:
            X["NUM_" + site] = X.apply(
                lambda x: self.count_site_nums(x["COMPOSITION"], x[site]), axis=1
            )
        return X


class PopulateElementalProp(BaseEstimator, TransformerMixin):
    def __init__(self, ep: pd.DataFrame, prefixes: List[str], sites: List[str]):
        self.ep = ep
        self.prefixes = prefixes
        self.sites = sites
        return None

    def fit(self, X: pd.DataFrame, y=None):
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        X = X.copy()

        X[self.sites] = X[self.sites].astype("str")

        for i in range(len(self.prefixes)):
            X = X.merge(
                self.ep.add_prefix(self.prefixes[i] + "_"),
                how="left",
                left_on=self.sites[i],
                right_on=self.prefixes[i] + "_SYMBOL",
            )

        return X


class ShannonRadius(BaseEstimator, TransformerMixin):
    def __init__(
        self,
        a_shan: pd.DataFrame,
        b_shan: pd.DataFrame,
        prefixes: List[str],
        sites: List[str],
    ):
        self.a_shan = a_shan
        self.b_shan = b_shan
        self.a_prefixes = prefixes[0:3]
        self.b_prefixes = prefixes[3:6]
        self.a_sites = sites[0:3]
        self.b_sites = sites[3:6]
        return None

    def fit(self, X: pd.DataFrame, y: pd.Series = None):
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        X = X.copy()

        for i in range(len(self.a_prefixes)):

            X = X.merge(
                self.a_shan.add_prefix(self.a_prefixes[i] + "_SHAN_"),
                how="left",
                left_on=self.a_sites[i],
                right_on=self.a_prefixes[i] + "_SHAN_A_SITE",
            )

        for i in range(len(self.b_prefixes)):
            X = X.merge(
                self.b_shan.add_prefix(self.b_prefixes[i] + "_SHAN_"),
                how="left",
                left_on=self.b_sites[i],
                right_on=self.b_prefixes[i] + "_SHAN_B_SITE",
            )

        cols_to_drop = [
            col
            for col in X.columns
            if ("_SHAN_A_SITE" in col) or ("_SHAN_B_SITE" in col)
        ]

        X = X.drop(columns=cols_to_drop)

        return X


class CompositionAveragedProperties(BaseEstimator, TransformerMixin):
    def __init__(self, sites: List[str], ep: pd.DataFrame, prefixes: List[str]):
        self.a_sites = sites[0:3]
        self.b_sites = sites[3:6]
        self.a_num_sites = ["NUM_" + site for site in sites[0:3]]
        self.b_num_sites = ["NUM_" + site for site in sites[3:6]]
        self.ep = ep
        self.prefixes = prefixes
        self.a_props = [""]
        self.b_props = [""]
        return None

    def fit(self, X: pd.DataFrame, y: pd.Series = None):
        return self

    def get_wt_avgd_property(
        self, prop: npt.ArrayLike, num_site: npt.ArrayLike
    ) -> float:
        return np.sum(np.multiply(prop, num_site)) / 8

    def get_max(self, sites, num_site):
        return sites[np.argmax(num_site)]

    def all_max(self, triplets: List[float]) -> float:
        return max(triplets)

    def all_min(self, triplets: List[float]) -> float:
        return min(triplets)

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        X = X.copy()

        for property in config.model_config.relevant_properties:

            property_list = [prefix + "_" + property for prefix in self.prefixes]

            if property not in ["SHAN_A_RADII", "SHAN_B_RADII"]:
                self.a_props = property_list[0:3]
                self.b_props = property_list[3:6]

                X["A_WT_AVG_" + property] = X.apply(
                    lambda x: self.get_wt_avgd_property(
                        x[self.a_props].fillna(0).to_numpy(),
                        x[self.a_num_sites].fillna(0).to_numpy(),
                    ),
                    axis=1,
                )

                X["A_ALL_MAX_" + property] = X.apply(
                    lambda x: self.all_max(list(x[self.a_props])), axis=1
                )

                X["A_ALL_MIN_" + property] = X.apply(
                    lambda x: self.all_min(list(x[self.a_props])), axis=1
                )

                X["B_WT_AVG_" + property] = X.apply(
                    lambda x: self.get_wt_avgd_property(
                        x[self.b_props].fillna(0).to_numpy(),
                        x[self.b_num_sites].fillna(0).to_numpy(),
                    ),
                    axis=1,
                )

                X["B_ALL_MAX_" + property] = X.apply(
                    lambda x: self.all_max(list(x[self.b_props])), axis=1
                )

                X["B_ALL_MIN_" + property] = X.apply(
                    lambda x: self.all_min(list(x[self.b_props])), axis=1
                )

                X["A_RANGE_" + property] = (
                    X["A_ALL_MAX_" + property] - X["A_ALL_MIN_" + property]
                )
                X["B_RANGE_" + property] = (
                    X["B_ALL_MAX_" + property] - X["B_ALL_MIN_" + property]
                )

            elif property == "SHAN_A_RADII":
                self.a_props = property_list[0:3]
                X["A_WT_AVG_SHAN_A_RADII"] = X.apply(
                    lambda x: self.get_wt_avgd_property(
                        x[self.a_props].fillna(0).to_numpy(),
                        x[self.a_num_sites].fillna(0).to_numpy(),
                    ),
                    axis=1,
                )

                X["A_ALL_MAX_" + property] = X.apply(
                    lambda x: self.all_max(list(x[self.a_props])), axis=1
                )
                X["A_ALL_MIN_" + property] = X.apply(
                    lambda x: self.all_min(list(x[self.a_props])), axis=1
                )
                X["A_RANGE_" + property] = (
                    X["A_ALL_MAX_" + property] - X["A_ALL_MIN_" + property]
                )

            elif property == "SHAN_B_RADII":
                self.b_props = property_list[3:6]
                X["B_WT_AVG_SHAN_B_RADII"] = X.apply(
                    lambda x: self.get_wt_avgd_property(
                        x[self.b_props].fillna(0).to_numpy(),
                        x[self.b_num_sites].fillna(0).to_numpy(),
                    ),
                    axis=1,
                )
                X["B_ALL_MAX_" + property] = X.apply(
                    lambda x: self.all_max(list(x[self.b_props])), axis=1
                )
                X["B_ALL_MIN_" + property] = X.apply(
                    lambda x: self.all_min(list(x[self.b_props])), axis=1
                )
                X["B_RANGE_" + property] = (
                    X["B_ALL_MAX_" + property] - X["B_ALL_MIN_" + property]
                )

        X["A_MAX"] = X.apply(
            lambda x: self.get_max(
                list(x[self.a_sites].fillna(np.nan)),
                list(x[self.a_num_sites].fillna(0)),
            ),
            axis=1,
        )
        X["B_MAX"] = X.apply(
            lambda x: self.get_max(
                list(x[self.b_sites].fillna(np.nan)),
                list(x[self.b_num_sites].fillna(0)),
            ),
            axis=1,
        )

        return X


class PopulateMajorityIonProperties(BaseEstimator, TransformerMixin):
    def __init__(self, ep: pd.DataFrame, cols: List[str]):
        self.ep = ep
        self.cols = cols
        return None

    def fit(self, X: pd.DataFrame, y: pd.Series = None):
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        X = X.copy()

        for col in self.cols:
            X = X.merge(
                self.ep.add_prefix(col + "_"),
                how="left",
                left_on=col,
                right_on=col + "_SYMBOL",
            )

        properties = self.ep.drop(columns=["SYMBOL"]).columns

        for property in properties:
            X["AVG_MAJ_" + property] = X[
                ["A_MAX_" + property, "B_MAX_" + property]
            ].mean(axis=1)
            X["DIFF_MAJ_" + property] = X["A_MAX_" + property] - X["B_MAX_" + property]

        return X


class KeepDesiredFeatures(BaseEstimator, TransformerMixin):
    def __init__(self, features: List[str]):
        self.features = features

    def fit(self, X: pd.DataFrame, y: pd.Series = None):
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        X = X.copy()
        return X[self.features].fillna(0)
